#!/usr/bin/perl

# Copyright (c) 1994 by Vixie Enterprises
# 
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies, and that
# the name of Vixie Enterprises not be used in advertising or publicity
# pertaining to distribution of the document or software without specific,
# written prior permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND Vixie Enterprises DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL Vixie Enterprises
# BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
# OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
#
# Author:       James A. Brister <brister@vix.com> -- berkeley-unix --
# Start Date:	Tue Oct 11 21:19:28 1994
# Project:	SENDS
# File:		$Id: sendsConfigLib.pl,v 1.2 1994/10/13 21:49:49 brister Exp $
# Description:	A modified version of the config package to allow for
#		merging of multiple entries (put here 'cause don't yet
#		know if changing the library will break things).
# 
# Update History:
#
# $Log: sendsConfigLib.pl,v $
# Revision 1.2  1994/10/13  21:49:49  brister
# More RFC1101 related changes.
#
# Revision 1.1  1994/10/11  12:04:48  brister
# Initial revision
#
#

## From original version:

#
# A fairly general config package. The main parsing routing will parse
# a config file and initialize an associative array based on the
# values found therein. Another array is passed in that defines the
# legal keywords in the config file. A facility for having the parser
# verify keyword values exists, and flags for making sure that all
# required keys were defined in the config file.
#
# The entry point is sub config'processConfig.
#
# Global variables are:
#
#	$config'commentChar	the character that defines comments.
#				(default is '#').
#	$config'Error		an error string that is appropriately set
#				if a routine fails. The Value is undefined
#				(in the general sense, NOT the Perl sense)
#				if a routing returns a successful value.
#
# Constants are:
#
#	$config'AllowExtras	passed in via the FLAGS argument if unknown
#				keywords found in the config file are
#				allowed to be there (and their values will
#				be returned with all the others keywords
#				not specified in the parameter list
#	$config'MustHaveExpr	passed in via the FLAGS argument if every
#				keyword found MUST be verifed. An
#				expression of key '*' can be used to catch
#				all unspecified keys.
#
# Keyword validity testing:
#
#	Each keyword/value pair extracted from the config file will be
#	tested against a perl expression supplied by the user. The
#	expression must evaluate to true for the keyword and the whole
#	config file to be accepted. The variable "$_" is set to the value
#	found in the config file before the evaluation is done. See the
#	exaple below for more detail.
#
#	The expression matching key '*' can be specified to match all keys
#	that don't have a specific matching expression.
#
#	Three semi-useful functions are supplied for verification purposes:
#
#		config'checkPosInt
#			Can optionally take a MIN and MAX argument
#			(defaults to 0 and MAXINT). returns TRUE if $_ is a
#			number between MIN and MAX
#
#		config'nonEmptyString
#			return TRUE if $_ is a non-empty string.
#
#		config'emptyString
#			returns TRUE if $_ is an empty string.
#
#
#	see the example below for their usage.
#

package config ;

$commentChar = "#" ;
$Error = '' ;

$colonRequired = 1 ;   # this should be configurable

$AllowExtras = (1 << 0) ;
$MustHaveExpr = (1 << 2) ;

# $debug = 1 ;


##############################################################################
##
## processConfig
##
## Parameters: (*returnValue, $configFileName, *requiredKeys,
##		*optionalKeys, *defaultValues, *verifyExprs, $flags)
##
##
## Process the file CONFIGFILENAME. RETURNVAL is the resulting assoc array,
## where each key is a keyword from the config file (tansformed to lower
## case), and its value is the value from the config file. REQUIREDKEYS is
## an array of keywords that MUST be present in the config
## file or must have a value in the DEFAULTVALUES associative array.
##  OPTIONALKEYS is a list of keys that MAY be in the config
## file. VERIFYEXPRS is an associatve array of keyword/perl-expression
## pairs. For every entry in VERIFYEXPRS the perl-expression will be tested
## against the found value. If the expression is FALSE, then processConfig
## fails. FLAGS is an integer that may be the OR'd value of the predefined
## constantrs ($AllMustExist and $AllowExtras--see description above)
##
##
## Return value:
##
##		1 on success
##		0 on failure
##
## Example:
##
##
##   @RequiredKeys = ('primary',
##   		 'name_servers',
##   		 'hostmaster',
##   		 'zone_master',
##   		 'zone_expire',
##   		 'zone_retry', 
##   		 'zone_refresh',
##   		 'zone_minimum',
##   		 'notify_addr',
##   		 'bad_domains',
##   );
##   
##   @OptionalKeys = ( 'debug', 'sendmail', 'tmpdir' ) ;
##   
##   %VerifyExprs = ('*', '$config\'Error = "$config\'key is an empty string", 
##                         &config\'nonEmptyString',    # catch-all test
##   	   	'primary', '$config\'Error = "primary is not a dotted quad",
##   			   /^(\d+)\.(\d+)\.(\d+)\.(\d+)/',
##   		'zone_expire', "&config'checkPosInt",
##   		'zone_retry', "&config'checkPosInt",
##   		'zone_refresh', "&config'checkPosInt",
##   		'zone_minimum', "&config'checkPosInt",
##   		'sendmail', '$config\'Error = "sendmail value is not valid",
##   			    -f $_ && -x _',
##   		'debug', "&config'checkPosInt",
##   		'tmpdir', '$config\'Error = "$config\'key is not a directory", -d ',
##   
##   );
##   
##   $flags = ($config'AllowExtras | $config'MustHaveExpr) ;
##   
##   %Config = () ;
##
##   %defaults = (
##   	'zone_expire', '604800',
##   	'zone_retry', '600',
##   	'zone_refresh', '1800',
##   	'zone_minimum', '1800',
##   	'tmpdir', '/var/tmp',
##   	'debug', '0',
##   );
##   
##   
##   die "SENDS config failed: $config'Error\n" 
##   	unless &config'processConfig (*Config,$configFileDefault,
##   				      *RequiredKeys, *OptionalKeys,
##   				      *defaults, *VerifyExprs, $flags) ;
##
##
##
##            
##
## The config file has a simple keyword/value structure, with whitespace
## separating value from the keyword. Continuation lines are allowed and must
## start with whitespace. (like Mail headers but without a mandatory ':')
## 

sub processConfig {
	local (*retVal,$configFile,*reqList,*optList,*defaults,
	       *exprs,$flags) = @_ ;
	local ($key,$value,$i) ;
	local ($required) = @reqList ;

	print STDERR "config'processConfig:($configFile, (",
			join (' ',@reqList),"), (",
			join (' ',@optList),"), (",
			join (' ',%exprs),"), $flags))\n" if $debug ;

	$Error = "No config file specified" ;

	return 0 unless $configFile;

	$Error = undef ;

	if ( !open (CF,$configFile) ) {
		$Error = "Cannot open config file: $configFile: $!" ;
		return 0 ;
	}

	print STDERR "config'processConfig: opened $configFile\n" if $debug ;

	while ( <CF> ) {
		chop ;
		print STDERR "$. $_\n" if $debug ;
		s/${commentChar}.*$// if defined ($commentChar) ;
		$Error = undef ;
		if ( /^\s*$/ ) {
			$key = undef ;
			next ;
		} elsif ( /^\s+/ && ! defined ($key) ) {
			$Error = "leading spaces on line # $. are illegal" ;
			close (CF) ;
			return 0 ;
		} elsif ( /^\s+(.*)$/ ) {
			if ( defined ($retVal{$key}) ) {
				# a continuation line
				$retVal{$key} .= " $1" ;
			} else {
				$retVal{$key} = $1 ;
			}
		} else {	
			($key,$value) = split (' ',$_,2) ;
			$key =~ tr/A-Z/a-z/ ;
			if  ($colonRequired && $key !~ /:/) {
				$Error = "Malformed line ($.): $_\n" ;
				return 0 ;
			}
			$key =~ s/:$// ;

			# we allow multiple lines with the same key and
			# merge their values.
			if ( defined ($retVal{$key}) ) {
				$retVal{$key} .= " " . $value ;
			} else {
				$retVal{$key} = $value ;
			}
		}
	}

	close (CF) ;

	foreach $key (keys (%retVal)) {
		return 0 unless &validateEntry ($key,
						$retVal{$key},
						*required, 
						*optList,
						$exprs{$key} || $exprs{'*'},
						$flags) ;
	}
	#
	# Now make sure all the required keywords were seen. If they 
        # weren't and if defaults were specified then set them.
	#

	print STDERR "Checking required keywords\n" if $debug ;
	local (%markerV) ;
	grep ($markerV{$_}++,keys %retVal) ;
	local (@diff) = grep(!$markerV{$_},@reqList) ; # (@reqList,@optList)) ;
	if ( @diff != 0 ) {
		local ($missingKeys) = "" ;
		foreach $i ( @diff ) {
			if ( defined $defaults{$i} ) {
				$retVal{$i} = $defaults{$i} ;
			} else { 
				$missingKeys .= " $i" ;
			}
		}

		if ( $missingKeys ne "" ) {
			$Error = "Missing keywords: $missingKeys" ;
			return 0 ;
		} else {
			return 1 ;
		}
	}

	1 ;
}


##############################################################################
##
## validateEntry
##
## verify that the key/value combination found in the config file is valid.
## The key must be on the required or optional lists, or the AllowExtras
## flag must be set. The given Perl expression if defined, must
## evaluate to TRUE also.
##
## Returns 1 on success and 0 on failure.
##

sub validateEntry {
	local ($key,$value,*requiredList,*optionalList,$expr,$flags) = @_ ;
	local ($where) ;

	print STDERR "Validating: \t$key, \n\t\t$value,\n\t\t$expr\n" 
		if $debug ;
		     
	#
	# Make sure the key word is on the required or optional lists. If
	# not then the flag AllowExtras better have been set.
	#
	undef $where ;
	for ($[ .. $#requiredList ) {
		$where = $_, last if $requiredList [$_] eq $key ;
	}
	if ( !defined ($where) ) {
		for ($[ .. $#optionalList ) {
			$where = $_, last if $optionalList [$_] eq $key ;
		}
		if ( ! defined ($where) && (($flags & $AllowExtras) == 0)) {
			$Error = "unrecognised keyword: $key" ;
			return 0 ;
		}
	}
	
	#
	# Now evaluate the perl expression on it if need be.
	#
	if (defined ($expr) && ($expr ne "")) {
		$Error = "keyword validation failed: [$key,$value,$expr]" ;
		local ($_) = $value ;
		local ($validateError) = '' ;
		if ( !eval ("package main ;" . $expr) ) {
			$Error .= ":$@ " if $@ ne "" ;
			if ( $validateError ne "" ) {
				$Error .= ": $@: $validateError" ;
			}
			return 0 ;
		}
	} elsif ($flags & $MustHaveExpr) {
		$Error = "keyword validation failed: [$key,$value]: " .
			"REQUIRED expression missing" ;
		return 0 ;
	}

	1 ;
}



##############################################################################
##                     Some Useful validation functions                     ##
##############################################################################


##############################################################################
##
## checkPosInt
##
## Make sure the value is a postitive integer.
##

sub checkPosInt {
	local ($minVal,$maxVal) = @_ ;

	$minVal = 0 unless defined ($minVal) ;
	$maxVal = 0x7fffffff unless defined ($maxVal) ;

	if ( /^0[xX]/ ) {
		if ( ! /^0[xX][a-f0-9]*/ ) { # include A-Z too?
			$validateError = "looks like a hex number but isn't" ;
			return 0 ;
		}
	} elsif ( ! /^\d+\s*$/ ) {
		$validateError = "not a number" ;
		return 0 ;
	}
	if ( $_ < $minVal ) {
		$validateError = "less than minimum ($_ vs $minVal)" ;
		return 0 ;
	}
	if ( $_ > $maxVal ) {
		$validateError = "greater than maximum ($_ vs $maxVal)" ;
		return 0 ;
	}

	1 ;
}


##############################################################################
##
## nonEmptyString
##
## Check for a non-empty string.
##

sub nonEmptyString {
	if ( $_ eq "" ) {
		$validateError = "empty string" ;
		return 0 ;
	}

	1;
}

sub emptyString {
	if ( $_ ne "" ) {
		$validateError = "non-empty string" ;
		return 0 ;
	}

	1 ;
}


1;



